# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A fake implementation for pathlib working with FakeFilesystem.
New in pyfakefs 3.0.

Usage:

* With fake_filesystem_unittest:
  If using fake_filesystem_unittest.TestCase, pathlib gets replaced
  by fake_pathlib together with other file system related modules.

* Stand-alone with FakeFilesystem:
  `filesystem = fake_filesystem.FakeFilesystem()`
  `fake_pathlib_module = fake_filesystem.FakePathlibModule(filesystem)`
  `path = fake_pathlib_module.Path('/foo/bar')`

Note: as the implementation is based on FakeFilesystem, all faked classes
(including PurePosixPath, PosixPath, PureWindowsPath and WindowsPath)
get the properties of the underlying fake filesystem.
"""

import errno
import fnmatch
import functools
import inspect
import ntpath
import os
import pathlib
import posixpath
import re
import sys
import warnings
from pathlib import PurePath
from typing import Callable, List, Optional
from urllib.parse import quote_from_bytes as urlquote_from_bytes

from pyfakefs import fake_scandir
from pyfakefs.fake_filesystem import FakeFilesystem
from pyfakefs.fake_open import fake_open
from pyfakefs.fake_os import FakeOsModule, use_original_os
from pyfakefs.fake_path import FakePathModule
from pyfakefs.helpers import IS_PYPY, is_called_from_skipped_module, FSType


_WIN_RESERVED_NAMES = (
    {"CON", "PRN", "AUX", "NUL"}
    | {"COM%d" % i for i in range(1, 10)}
    | {"LPT%d" % i for i in range(1, 10)}
)


def init_module(filesystem):
    """Initializes the fake module with the fake file system."""
    # pylint: disable=protected-access
    FakePath.filesystem = filesystem
    if sys.version_info < (3, 12):
        FakePathlibModule.WindowsPath._flavour = _FakeWindowsFlavour(filesystem)
        FakePathlibModule.PosixPath._flavour = _FakePosixFlavour(filesystem)

        # Pure POSIX path separators must be filesystem-independent.
        fake_pure_posix_flavour = _FakePosixFlavour(filesystem)
        fake_pure_posix_flavour.sep = "/"
        fake_pure_posix_flavour.altsep = None
        FakePathlibModule.PurePosixPath._flavour = fake_pure_posix_flavour

        # Pure Windows path separators must be filesystem-independent.
        fake_pure_nt_flavour = _FakeWindowsFlavour(filesystem)
        fake_pure_nt_flavour.sep = "\\"
        fake_pure_nt_flavour.altsep = "/"
        FakePathlibModule.PureWindowsPath._flavour = fake_pure_nt_flavour
    else:
        # in Python > 3.11, the flavour is no longer a separate class,
        # but points to the os-specific path module (posixpath/ntpath)
        fake_os_posix = FakeOsModule(filesystem)
        if filesystem.is_windows_fs:
            fake_os_posix.path = FakePosixPathModule(filesystem, fake_os_posix)
        fake_os_windows = FakeOsModule(filesystem)
        if not filesystem.is_windows_fs:
            fake_os_windows.path = FakeWindowsPathModule(filesystem, fake_os_windows)

        parser_name = "_flavour" if sys.version_info < (3, 13) else "parser"

        # Pure POSIX path properties must be filesystem independent.
        setattr(FakePathlibModule.PurePosixPath, parser_name, fake_os_posix.path)

        # Pure Windows path properties must be filesystem independent.
        setattr(FakePathlibModule.PureWindowsPath, parser_name, fake_os_windows.path)


def _wrap_strfunc(fake_fct, original_fct):
    @functools.wraps(fake_fct)
    def _wrapped(pathobj, *args, **kwargs):
        fs: FakeFilesystem = pathobj.filesystem
        if fs.patcher:
            if is_called_from_skipped_module(
                skip_names=fs.patcher.skip_names,
                case_sensitive=fs.is_case_sensitive,
            ):
                return original_fct(str(pathobj), *args, **kwargs)
        return fake_fct(fs, str(pathobj), *args, **kwargs)

    return staticmethod(_wrapped)


def _wrap_binary_strfunc(fake_fct, original_fct):
    @functools.wraps(fake_fct)
    def _wrapped(pathobj1, pathobj2, *args):
        fs: FakeFilesystem = pathobj1.filesystem
        if fs.patcher:
            if is_called_from_skipped_module(
                skip_names=fs.patcher.skip_names,
                case_sensitive=fs.is_case_sensitive,
            ):
                return original_fct(str(pathobj1), str(pathobj2), *args)
        return fake_fct(fs, str(pathobj1), str(pathobj2), *args)

    return staticmethod(_wrapped)


def _wrap_binary_strfunc_reverse(fake_fct, original_fct):
    @functools.wraps(fake_fct)
    def _wrapped(pathobj1, pathobj2, *args):
        fs: FakeFilesystem = pathobj2.filesystem
        if fs.patcher:
            if is_called_from_skipped_module(
                skip_names=fs.patcher.skip_names,
                case_sensitive=fs.is_case_sensitive,
            ):
                return original_fct(str(pathobj2), str(pathobj1), *args)
        return fake_fct(fs, str(pathobj2), str(pathobj1), *args)

    return staticmethod(_wrapped)


try:
    accessor = pathlib._Accessor  # type: ignore[attr-defined]
except AttributeError:
    accessor = object


class _FakeAccessor(accessor):  # type: ignore[valid-type, misc]
    """Accessor which forwards some of the functions to FakeFilesystem
    methods.
    """

    stat = _wrap_strfunc(FakeFilesystem.stat, os.stat)

    lstat = _wrap_strfunc(
        lambda fs, path: FakeFilesystem.stat(fs, path, follow_symlinks=False), os.lstat
    )

    listdir = _wrap_strfunc(FakeFilesystem.listdir, os.listdir)
    scandir = _wrap_strfunc(fake_scandir.scandir, os.scandir)

    if hasattr(os, "lchmod"):
        lchmod = _wrap_strfunc(
            lambda fs, path, mode: FakeFilesystem.chmod(
                fs, path, mode, follow_symlinks=False
            ),
            os.lchmod,
        )
    else:

        def lchmod(self, pathobj, *args, **kwargs):
            """Raises not implemented for Windows systems."""
            raise NotImplementedError("lchmod() not available on this system")

    def chmod(self, pathobj, *args, **kwargs):
        if "follow_symlinks" in kwargs:
            if sys.version_info < (3, 10):
                raise TypeError(
                    "chmod() got an unexpected keyword " "argument 'follow_symlinks'"
                )

            if not kwargs["follow_symlinks"] and (
                os.chmod not in os.supports_follow_symlinks or IS_PYPY
            ):
                raise NotImplementedError(
                    "`follow_symlinks` for chmod() is not available " "on this system"
                )
        return pathobj.filesystem.chmod(str(pathobj), *args, **kwargs)

    mkdir = _wrap_strfunc(FakeFilesystem.makedir, os.mkdir)

    unlink = _wrap_strfunc(FakeFilesystem.remove, os.unlink)

    rmdir = _wrap_strfunc(FakeFilesystem.rmdir, os.rmdir)

    rename = _wrap_binary_strfunc(FakeFilesystem.rename, os.rename)

    replace = _wrap_binary_strfunc(
        lambda fs, old_path, new_path: FakeFilesystem.rename(
            fs, old_path, new_path, force_replace=True
        ),
        os.replace,
    )

    symlink = _wrap_binary_strfunc_reverse(
        lambda fs, fpath, target, target_is_dir: FakeFilesystem.create_symlink(
            fs, fpath, target, create_missing_dirs=False
        ),
        os.symlink,
    )

    if (3, 8) <= sys.version_info:
        link_to = _wrap_binary_strfunc(
            lambda fs, file_path, link_target: FakeFilesystem.link(
                fs, file_path, link_target
            ),
            os.link,
        )

    if sys.version_info >= (3, 10):
        link = _wrap_binary_strfunc(
            lambda fs, file_path, link_target: FakeFilesystem.link(
                fs, file_path, link_target
            ),
            os.link,
        )

        # this will use the fake filesystem because os is patched
        def getcwd(self):
            return os.getcwd()

    readlink = _wrap_strfunc(FakeFilesystem.readlink, os.readlink)

    utime = _wrap_strfunc(FakeFilesystem.utime, os.utime)


_fake_accessor = _FakeAccessor()

if sys.version_info < (3, 12):
    flavour = pathlib._Flavour  # type: ignore[attr-defined]

    class _FakeFlavour(flavour):  # type: ignore[valid-type, misc]
        """Fake Flavour implementation used by PurePath and _Flavour"""

        filesystem = None

        ext_namespace_prefix = "\\\\?\\"

        drive_letters = {chr(x) for x in range(ord("a"), ord("z") + 1)} | {
            chr(x) for x in range(ord("A"), ord("Z") + 1)
        }

        def __init__(self, filesystem):
            self.filesystem = filesystem
            super().__init__()

        @staticmethod
        def _split_extended_path(path, ext_prefix=ext_namespace_prefix):
            prefix = ""
            if path.startswith(ext_prefix):
                prefix = path[:4]
                path = path[4:]
                if path.startswith("UNC\\"):
                    prefix += path[:3]
                    path = "\\" + path[3:]
            return prefix, path

        def _splitroot_with_drive(self, path, sep):
            first = path[0:1]
            second = path[1:2]
            if second == sep and first == sep:
                # extended paths should also disable the collapsing of "."
                # components (according to MSDN docs).
                prefix, path = self._split_extended_path(path)
                first = path[0:1]
                second = path[1:2]
            else:
                prefix = ""
            third = path[2:3]
            if second == sep and first == sep and third != sep:
                # is a UNC path:
                # vvvvvvvvvvvvvvvvvvvvv root
                # \\machine\mountpoint\directory\etc\...
                #            directory ^^^^^^^^^^^^^^
                index = path.find(sep, 2)
                if index != -1:
                    index2 = path.find(sep, index + 1)
                    # a UNC path can't have two slashes in a row
                    # (after the initial two)
                    if index2 != index + 1:
                        if index2 == -1:
                            index2 = len(path)
                        if prefix:
                            return prefix + path[1:index2], sep, path[index2 + 1 :]
                        return path[:index2], sep, path[index2 + 1 :]
            drv = root = ""
            if second == ":" and first in self.drive_letters:
                drv = path[:2]
                path = path[2:]
                first = third
            if first == sep:
                root = first
                path = path.lstrip(sep)
            return prefix + drv, root, path

        @staticmethod
        def _splitroot_posix(path, sep):
            if path and path[0] == sep:
                stripped_part = path.lstrip(sep)
                if len(path) - len(stripped_part) == 2:
                    return "", sep * 2, stripped_part
                return "", sep, stripped_part
            else:
                return "", "", path

        def splitroot(self, path, sep=None):
            """Split path into drive, root and rest."""
            is_windows = isinstance(self, _FakeWindowsFlavour)
            if sep is None:
                if is_windows == self.filesystem.is_windows_fs:
                    sep = self.filesystem.path_separator
                else:
                    sep = self.sep
            if is_windows:
                return self._splitroot_with_drive(path, sep)
            return self._splitroot_posix(path, sep)

        def casefold(self, path):
            """Return the lower-case version of s for a Windows filesystem."""
            if self.filesystem.is_windows_fs:
                return path.lower()
            return path

        def casefold_parts(self, parts):
            """Return the lower-case version of parts for a Windows filesystem."""
            if self.filesystem.is_windows_fs:
                return [p.lower() for p in parts]
            return parts

        def _resolve_posix(self, path, strict):
            sep = self.sep
            seen = {}

            def _resolve(path, rest):
                if rest.startswith(sep):
                    path = ""

                for name in rest.split(sep):
                    if not name or name == ".":
                        # current dir
                        continue
                    if name == "..":
                        # parent dir
                        path, _, _ = path.rpartition(sep)
                        continue
                    newpath = path + sep + name
                    if newpath in seen:
                        # Already seen this path
                        path = seen[newpath]
                        if path is not None:
                            # use cached value
                            continue
                        # The symlink is not resolved, so we must have
                        # a symlink loop.
                        raise RuntimeError("Symlink loop from %r" % newpath)
                    # Resolve the symbolic link
                    try:
                        target = self.filesystem.readlink(newpath)
                    except OSError as e:
                        if e.errno != errno.EINVAL and strict:
                            raise
                        # Not a symlink, or non-strict mode. We just leave the path
                        # untouched.
                        path = newpath
                    else:
                        seen[newpath] = None  # not resolved symlink
                        path = _resolve(path, target)
                        seen[newpath] = path  # resolved symlink

                return path

            # NOTE: according to POSIX, getcwd() cannot contain path components
            # which are symlinks.
            base = "" if path.is_absolute() else self.filesystem.cwd
            return _resolve(base, str(path)) or sep

        def _resolve_windows(self, path, strict):
            path = str(path)
            if not path:
                return os.getcwd()
            previous_s = None
            if strict:
                if not self.filesystem.exists(path):
                    self.filesystem.raise_os_error(errno.ENOENT, path)
                return self.filesystem.resolve_path(path)
            else:
                while True:
                    try:
                        path = self.filesystem.resolve_path(path)
                    except OSError:
                        previous_s = path
                        path = self.filesystem.splitpath(path)[0]
                    else:
                        if previous_s is None:
                            return path
                        return self.filesystem.joinpaths(
                            path, os.path.basename(previous_s)
                        )

        def resolve(self, path, strict):
            """Make the path absolute, resolving any symlinks."""
            if self.filesystem.is_windows_fs:
                return self._resolve_windows(path, strict)
            return self._resolve_posix(path, strict)

        def gethomedir(self, username):
            """Return the home directory of the current user."""
            if not username:
                try:
                    return os.environ["HOME"]
                except KeyError:
                    import pwd

                    return pwd.getpwuid(os.getuid()).pw_dir
            else:
                import pwd

                try:
                    return pwd.getpwnam(username).pw_dir
                except KeyError:
                    raise RuntimeError(
                        "Can't determine home directory " "for %r" % username
                    )

    class _FakeWindowsFlavour(_FakeFlavour):
        """Flavour used by PureWindowsPath with some Windows specific
        implementations independent of FakeFilesystem properties.
        """

        sep = "\\"
        altsep = "/"
        has_drv = True
        pathmod = ntpath

        def is_reserved(self, parts):
            """Return True if the path is considered reserved under Windows."""

            # NOTE: the rules for reserved names seem somewhat complicated
            # (e.g. r"..\NUL" is reserved but not r"foo\NUL").
            # We err on the side of caution and return True for paths which are
            # not considered reserved by Windows.
            if not parts:
                return False
            if self.filesystem.is_windows_fs and parts[0].startswith("\\\\"):
                # UNC paths are never reserved
                return False
            return parts[-1].partition(".")[0].upper() in _WIN_RESERVED_NAMES

        def make_uri(self, path):
            """Return a file URI for the given path"""

            # Under Windows, file URIs use the UTF-8 encoding.
            # original version, not faked
            drive = path.drive
            if len(drive) == 2 and drive[1] == ":":
                # It's a path on a local drive => 'file:///c:/a/b'
                rest = path.as_posix()[2:].lstrip("/")
                return "file:///{}/{}".format(
                    drive,
                    urlquote_from_bytes(rest.encode("utf-8")),
                )
            else:
                # It's a path on a network drive => 'file://host/share/a/b'
                return "file:" + urlquote_from_bytes(path.as_posix().encode("utf-8"))

        def gethomedir(self, username):
            """Return the home directory of the current user."""

            # original version, not faked
            if "HOME" in os.environ:
                userhome = os.environ["HOME"]
            elif "USERPROFILE" in os.environ:
                userhome = os.environ["USERPROFILE"]
            elif "HOMEPATH" in os.environ:
                try:
                    drv = os.environ["HOMEDRIVE"]
                except KeyError:
                    drv = ""
                userhome = drv + os.environ["HOMEPATH"]
            else:
                raise RuntimeError("Can't determine home directory")

            if username:
                # Try to guess user home directory.  By default all users
                # directories are located in the same place and are named by
                # corresponding usernames.  If current user home directory points
                # to nonstandard place, this guess is likely wrong.
                if os.environ["USERNAME"] != username:
                    drv, root, parts = self.parse_parts((userhome,))
                    if parts[-1] != os.environ["USERNAME"]:
                        raise RuntimeError(
                            "Can't determine home directory " "for %r" % username
                        )
                    parts[-1] = username
                    if drv or root:
                        userhome = drv + root + self.join(parts[1:])
                    else:
                        userhome = self.join(parts)
            return userhome

        def compile_pattern(self, pattern):
            return re.compile(fnmatch.translate(pattern), re.IGNORECASE).fullmatch

    class _FakePosixFlavour(_FakeFlavour):
        """Flavour used by PurePosixPath with some Unix specific implementations
        independent of FakeFilesystem properties.
        """

        sep = "/"
        altsep: Optional[str] = None
        has_drv = False
        pathmod = posixpath

        def is_reserved(self, parts):
            return False

        def make_uri(self, path):
            # We represent the path using the local filesystem encoding,
            # for portability to other applications.
            bpath = bytes(path)
            return "file://" + urlquote_from_bytes(bpath)

        def gethomedir(self, username):
            # original version, not faked
            if not username:
                try:
                    return os.environ["HOME"]
                except KeyError:
                    import pwd

                    return pwd.getpwuid(os.getuid()).pw_dir
            else:
                import pwd

                try:
                    return pwd.getpwnam(username).pw_dir
                except KeyError:
                    raise RuntimeError(
                        "Can't determine home directory " "for %r" % username
                    )

        def compile_pattern(self, pattern):
            return re.compile(fnmatch.translate(pattern)).fullmatch
else:  # Python >= 3.12

    class FakePosixPathModule(FakePathModule):
        def __init__(self, filesystem: "FakeFilesystem", os_module: "FakeOsModule"):
            super().__init__(filesystem, os_module)
            with self.filesystem.use_fs_type(FSType.POSIX):
                self.reset(self.filesystem)

    class FakeWindowsPathModule(FakePathModule):
        def __init__(self, filesystem: "FakeFilesystem", os_module: "FakeOsModule"):
            super().__init__(filesystem, os_module)
            with self.filesystem.use_fs_type(FSType.WINDOWS):
                self.reset(self.filesystem)

    def with_fs_type(f: Callable, fs_type: FSType) -> Callable:
        """Decorator used for fake_path methods to ensure that
        the correct filesystem type is used."""

        @functools.wraps(f)
        def wrapped(self, *args, **kwargs):
            with self.filesystem.use_fs_type(fs_type):
                return f(self, *args, **kwargs)

        return wrapped

    # decorate all public functions to use the correct fs type
    for fct_name in FakePathModule.dir():
        fn = getattr(FakePathModule, fct_name)
        setattr(FakeWindowsPathModule, fct_name, with_fs_type(fn, FSType.WINDOWS))
        setattr(FakePosixPathModule, fct_name, with_fs_type(fn, FSType.POSIX))


class FakePath(pathlib.Path):
    """Replacement for pathlib.Path. Reimplement some methods to use
    fake filesystem. The rest of the methods work as they are, as they will
    use the fake accessor.
    New in pyfakefs 3.0.
    """

    # the underlying fake filesystem
    filesystem = None
    skip_names: List[str] = []

    def __new__(cls, *args, **kwargs):
        """Creates the correct subclass based on OS."""
        if cls is FakePathlibModule.Path:
            cls = (
                FakePathlibModule.WindowsPath
                if cls.filesystem.is_windows_fs
                else FakePathlibModule.PosixPath
            )
        if sys.version_info < (3, 12):
            return cls._from_parts(args)  # pytype: disable=attribute-error
        else:
            return object.__new__(cls)

    if sys.version_info[:2] == (3, 10):
        # Overwritten class methods to call _init to set the fake accessor,
        # which is not done in Python 3.10, and not needed from Python 3.11 on
        @classmethod
        def _from_parts(cls, args):
            self = object.__new__(cls)
            self._init()
            drv, root, parts = self._parse_args(args)  # pytype: disable=attribute-error
            self._drv = drv
            self._root = root
            self._parts = parts
            return self

        @classmethod
        def _from_parsed_parts(cls, drv, root, parts):
            self = object.__new__(cls)
            self._drv = drv
            self._root = root
            self._parts = parts
            self._init()
            return self

    if sys.version_info < (3, 11):

        def _init(self, template=None):
            """Initializer called from base class."""
            # only needed until Python 3.10
            self._accessor = _fake_accessor
            # only needed until Python 3.8
            self._closed = False

    def _path(self):
        """Returns the underlying path string as used by the fake
        filesystem.
        """
        return str(self)

    @classmethod
    def cwd(cls):
        """Return a new path pointing to the current working directory
        (as returned by os.getcwd()).
        """
        return cls(cls.filesystem.cwd)

    if sys.version_info < (3, 12):  # in 3.12, we can use the pathlib implementation

        def resolve(self, strict=None):
            """Make the path absolute, resolving all symlinks on the way and also
            normalizing it (for example turning slashes into backslashes
            under Windows).

            Args:
                strict: If False (default) no exception is raised if the path
                    does not exist.

            Raises:
                OSError: if the path doesn't exist (strict=True)
            """
            if strict is None:
                strict = False
            self._raise_on_closed()
            path = self._flavour.resolve(
                self, strict=strict
            )  # pytype: disable=attribute-error
            if path is None:
                self.stat()
                path = str(self.absolute())
            path = self.filesystem.absnormpath(path)
            return FakePath(path)

    def open(self, mode="r", buffering=-1, encoding=None, errors=None, newline=None):
        """Open the file pointed by this path and return a fake file object.

        Raises:
            OSError: if the target object is a directory, the path is invalid
                or permission is denied.
        """
        self._raise_on_closed()
        return fake_open(
            self.filesystem,
            self.skip_names,
            self._path(),
            mode,
            buffering,
            encoding,
            errors,
            newline,
        )

    def read_bytes(self):
        """Open the fake file in bytes mode, read it, and close the file.

        Raises:
            OSError: if the target object is a directory, the path is
                invalid or permission is denied.
        """
        with fake_open(
            self.filesystem,
            self.skip_names,
            self._path(),
            mode="rb",
        ) as f:
            return f.read()

    def read_text(self, encoding=None, errors=None):
        """
        Open the fake file in text mode, read it, and close the file.
        """
        with fake_open(
            self.filesystem,
            self.skip_names,
            self._path(),
            mode="r",
            encoding=encoding,
            errors=errors,
        ) as f:
            return f.read()

    def write_bytes(self, data):
        """Open the fake file in bytes mode, write to it, and close the file.
        Args:
            data: the bytes to be written
        Raises:
            OSError: if the target object is a directory, the path is
                invalid or permission is denied.
        """
        # type-check for the buffer interface before truncating the file
        view = memoryview(data)
        with fake_open(
            self.filesystem,
            self.skip_names,
            self._path(),
            mode="wb",
        ) as f:
            return f.write(view)

    def write_text(self, data, encoding=None, errors=None, newline=None):
        """Open the fake file in text mode, write to it, and close
        the file.

        Args:
            data: the string to be written
            encoding: the encoding used for the string; if not given, the
                default locale encoding is used
            errors: (str) Defines how encoding errors are handled.
            newline: Controls universal newlines, passed to stream object.
                New in Python 3.10.
        Raises:
            TypeError: if data is not of type 'str'.
            OSError: if the target object is a directory, the path is
                invalid or permission is denied.
        """
        if not isinstance(data, str):
            raise TypeError("data must be str, not %s" % data.__class__.__name__)
        if newline is not None and sys.version_info < (3, 10):
            raise TypeError(
                "write_text() got an unexpected " "keyword argument 'newline'"
            )
        with fake_open(
            self.filesystem,
            self.skip_names,
            self._path(),
            mode="w",
            encoding=encoding,
            errors=errors,
            newline=newline,
        ) as f:
            return f.write(data)

    @classmethod
    def home(cls):
        """Return a new path pointing to the user's home directory (as
        returned by os.path.expanduser('~')).
        """
        home = os.path.expanduser("~")
        if cls.filesystem.is_windows_fs != (os.name == "nt"):
            username = os.path.split(home)[1]
            if cls.filesystem.is_windows_fs:
                home = os.path.join("C:", "Users", username)
            else:
                home = os.path.join("home", username)
            if not cls.filesystem.exists(home):
                cls.filesystem.create_dir(home)
        return cls(home.replace(os.sep, cls.filesystem.path_separator))

    def samefile(self, other_path):
        """Return whether other_path is the same or not as this file
        (as returned by os.path.samefile()).

        Args:
            other_path: A path object or string of the file object
            to be compared with

        Raises:
            OSError: if the filesystem object doesn't exist.
        """
        st = self.stat()
        try:
            other_st = other_path.stat()
        except AttributeError:
            other_st = self.filesystem.stat(other_path)
        return st.st_ino == other_st.st_ino and st.st_dev == other_st.st_dev

    def expanduser(self):
        """Return a new path with expanded ~ and ~user constructs
        (as returned by os.path.expanduser)
        """
        return FakePath(
            os.path.expanduser(self._path()).replace(
                os.path.sep, self.filesystem.path_separator
            )
        )

    def _raise_on_closed(self):
        if sys.version_info < (3, 9) and self._closed:
            self._raise_closed()

    def touch(self, mode=0o666, exist_ok=True):
        """Create a fake file for the path with the given access mode,
        if it doesn't exist.

        Args:
            mode: the file mode for the file if it does not exist
            exist_ok: if the file already exists and this is True, nothing
                happens, otherwise FileExistError is raised

        Raises:
            FileExistsError: if the file exists and exits_ok is False.
        """
        self._raise_on_closed()
        if self.exists():
            if exist_ok:
                self.filesystem.utime(self._path(), times=None)
            else:
                self.filesystem.raise_os_error(errno.EEXIST, self._path())
        else:
            fake_file = self.open("w", encoding="utf8")
            fake_file.close()
            self.chmod(mode)


def _warn_is_reserved_deprecated():
    if sys.version_info >= (3, 13):
        warnings.warn(
            "pathlib.PurePath.is_reserved() is deprecated and scheduled "
            "for removal in Python 3.15. Use os.path.isreserved() to detect "
            "reserved paths on Windows.",
            DeprecationWarning,
        )


class FakePathlibModule:
    """Uses FakeFilesystem to provide a fake pathlib module replacement.

    You need a fake_filesystem to use this:
    `filesystem = fake_filesystem.FakeFilesystem()`
    `fake_pathlib_module = fake_filesystem.FakePathlibModule(filesystem)`
    """

    def __init__(self, filesystem):
        """
        Initializes the module with the given filesystem.

        Args:
            filesystem: FakeFilesystem used to provide file system information
        """
        init_module(filesystem)
        self._pathlib_module = pathlib

    class PurePosixPath(PurePath):
        """A subclass of PurePath, that represents non-Windows filesystem
        paths"""

        __slots__ = ()
        if sys.version_info >= (3, 12):

            def is_reserved(self):
                _warn_is_reserved_deprecated()
                return False

            def is_absolute(self):
                with os.path.filesystem.use_fs_type(FSType.POSIX):  # type: ignore[module-attr]
                    return os.path.isabs(self)

            def joinpath(self, *pathsegments):
                with os.path.filesystem.use_fs_type(FSType.POSIX):  # type: ignore[module-attr]
                    return super().joinpath(*pathsegments)

    class PureWindowsPath(PurePath):
        """A subclass of PurePath, that represents Windows filesystem paths"""

        __slots__ = ()

        if sys.version_info >= (3, 12):
            """These are reimplemented because the PurePath implementation
            checks the flavour against ntpath/posixpath.
            """

            def is_reserved(self):
                _warn_is_reserved_deprecated()
                if sys.version_info < (3, 13):
                    if not self._tail or self._tail[0].startswith("\\\\"):
                        # UNC paths are never reserved.
                        return False
                    name = (
                        self._tail[-1].partition(".")[0].partition(":")[0].rstrip(" ")
                    )
                    return name.upper() in _WIN_RESERVED_NAMES
                with os.path.filesystem.use_fs_type(FSType.WINDOWS):  # type: ignore[module-attr]
                    return os.path.isreserved(self)

            def is_absolute(self):
                with os.path.filesystem.use_fs_type(FSType.WINDOWS):
                    return bool(self.drive and self.root)

    class WindowsPath(FakePath, PureWindowsPath):
        """A subclass of Path and PureWindowsPath that represents
        concrete Windows filesystem paths.
        """

        __slots__ = ()

        def owner(self):
            raise NotImplementedError("Path.owner() is unsupported on this system")

        def group(self):
            raise NotImplementedError("Path.group() is unsupported on this system")

        def is_mount(self):
            raise NotImplementedError("Path.is_mount() is unsupported on this system")

    class PosixPath(FakePath, PurePosixPath):
        """A subclass of Path and PurePosixPath that represents
        concrete non-Windows filesystem paths.
        """

        __slots__ = ()

        def owner(self):
            """Return the username of the file owner.
            It is assumed that `st_uid` is related to a real user,
            otherwise `KeyError` is raised.
            """
            import pwd

            return pwd.getpwuid(self.stat().st_uid).pw_name

        def group(self):
            """Return the group name of the file group.
            It is assumed that `st_gid` is related to a real group,
            otherwise `KeyError` is raised.
            """
            import grp

            return grp.getgrgid(self.stat().st_gid).gr_name

    Path = FakePath

    def __getattr__(self, name):
        """Forwards any unfaked calls to the standard pathlib module."""
        return getattr(self._pathlib_module, name)


class FakePathlibPathModule:
    """Patches `pathlib.Path` by passing all calls to FakePathlibModule."""

    fake_pathlib = None

    def __init__(self, filesystem=None):
        if self.fake_pathlib is None:
            self.__class__.fake_pathlib = FakePathlibModule(filesystem)

    @property
    def skip_names(self):
        return []  # not used, here to allow a setter

    @skip_names.setter
    def skip_names(self, value):
        # this is set from the patcher and passed to the fake Path class
        self.fake_pathlib.Path.skip_names = value

    def __call__(self, *args, **kwargs):
        return self.fake_pathlib.Path(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self.fake_pathlib.Path, name)

    @classmethod
    def __instancecheck__(cls, instance):
        # fake the inheritance to pass isinstance checks - see #666
        return isinstance(instance, PurePath)


class RealPath(pathlib.Path):
    """Replacement for `pathlib.Path` if it shall not be faked.
    Needed because `Path` in `pathlib` is always faked, even if `pathlib`
    itself is not.
    """

    if sys.version_info < (3, 12):
        _flavour = (
            pathlib._WindowsFlavour()  # type:ignore
            if os.name == "nt"
            else pathlib._PosixFlavour()  # type:ignore
        )  # type:ignore
    elif sys.version_info < (3, 13):
        _flavour = ntpath if os.name == "nt" else posixpath
    else:
        parser = ntpath if os.name == "nt" else posixpath

    def __new__(cls, *args, **kwargs):
        """Creates the correct subclass based on OS."""
        if cls is RealPathlibModule.Path:
            cls = (
                RealPathlibModule.WindowsPath  # pytype: disable=attribute-error
                if os.name == "nt"
                else RealPathlibModule.PosixPath  # pytype: disable=attribute-error
            )
        if sys.version_info < (3, 12):
            return cls._from_parts(args)  # pytype: disable=attribute-error
        else:
            return object.__new__(cls)


if sys.version_info > (3, 10):

    def with_original_os(f: Callable) -> Callable:
        """Decorator used for real pathlib Path methods to ensure that
        real os functions instead of faked ones are used."""

        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            with use_original_os():
                return f(*args, **kwargs)

        return wrapped

    for fct_name, fn in inspect.getmembers(RealPath, inspect.isfunction):
        if not fct_name.startswith("__"):
            setattr(RealPath, fct_name, with_original_os(fn))


class RealPathlibPathModule:
    """Patches `pathlib.Path` by passing all calls to RealPathlibModule."""

    real_pathlib = None

    @classmethod
    def __instancecheck__(cls, instance):
        # as we cannot derive from pathlib.Path, we fake
        # the inheritance to pass isinstance checks - see #666
        return isinstance(instance, PurePath)

    def __init__(self):
        if self.real_pathlib is None:
            self.__class__.real_pathlib = RealPathlibModule()

    def __call__(self, *args, **kwargs):
        return RealPath(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self.real_pathlib.Path, name)


class RealPathlibModule:
    """Used to replace `pathlib` for skipped modules.
    As the original `pathlib` is always patched to use the fake path,
    we need to provide a version which does not do this.
    """

    def __init__(self):
        self._pathlib_module = pathlib

    class PurePosixPath(PurePath):
        """A subclass of PurePath, that represents Posix filesystem paths"""

        __slots__ = ()

    class PureWindowsPath(PurePath):
        """A subclass of PurePath, that represents Windows filesystem paths"""

        __slots__ = ()

    if sys.platform == "win32":

        class WindowsPath(RealPath, PureWindowsPath):
            """A subclass of Path and PureWindowsPath that represents
            concrete Windows filesystem paths.
            """

            __slots__ = ()

    else:

        class PosixPath(RealPath, PurePosixPath):
            """A subclass of Path and PurePosixPath that represents
            concrete non-Windows filesystem paths.
            """

            __slots__ = ()

    Path = RealPath

    def __getattr__(self, name):
        """Forwards any unfaked calls to the standard pathlib module."""
        return getattr(self._pathlib_module, name)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A test suite that runs all tests for pyfakefs at once.
Excludes tests using external scandir package."""

import sys
import unittest

from pyfakefs import legacy_packages

legacy_packages.scandir = None
legacy_packages.pathlib2 = None

from pyfakefs.tests.all_tests import AllTests  # noqa: E402

if __name__ == "__main__":
    result = unittest.TextTestRunner(verbosity=2).run(AllTests().suite())
    sys.exit(int(not result.wasSuccessful()))



# Copyright 2017 John McGehee
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module is for testing pyfakefs
:py:class:`fake_filesystem_unittest.Patcher`.  It defines attributes that have
the same names as file modules, sudh as 'io` and `path`.  Since these are not
modules, :py:class:`fake_filesystem_unittest.Patcher` should not patch them.

Whenever a new module is added to
:py:meth:`fake_filesystem_unittest.Patcher._findModules`, the corresponding
attribute should be added here and in the test
:py:class:`fake_filesystem_unittest_test.TestAttributesWithFakeModuleNames`.
"""

os = "os attribute value"
path = "path attribute value"
pathlib = "pathlib attribute value"
shutil = "shutil attribute value"
io = "io attribute value"



# Copyright 2009 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for `fake_filesystem_shutil` if used in
`fake_filesystem_unittest.TestCase`.
Note that almost all of the functionality is delegated to the real `shutil`
and works correctly with the fake filesystem because of the faked `os` module.
"""

import os
import shutil
import sys
import tempfile
import unittest
from pathlib import Path

from pyfakefs import fake_filesystem_unittest
from pyfakefs.helpers import get_uid, set_uid, is_root, IS_PYPY
from pyfakefs.tests.test_utils import RealFsTestMixin

is_windows = sys.platform == "win32"


class RealFsTestCase(fake_filesystem_unittest.TestCase, RealFsTestMixin):
    def __init__(self, methodName="runTest"):
        fake_filesystem_unittest.TestCase.__init__(self, methodName)
        RealFsTestMixin.__init__(self)

    def setUp(self):
        RealFsTestMixin.setUp(self)
        self.cwd = os.getcwd()
        self.uid = get_uid()
        set_uid(1000)
        if not self.use_real_fs():
            self.setUpPyfakefs()
            self.filesystem = self.fs
            self.os = os
            self.open = open
            self.create_basepath()
            self.fs.set_disk_usage(1000, self.base_path)

    def tearDown(self):
        set_uid(self.uid)
        RealFsTestMixin.tearDown(self)

    @property
    def is_windows_fs(self):
        if self.use_real_fs():
            return sys.platform == "win32"
        return self.filesystem.is_windows_fs


class FakeShutilModuleTest(RealFsTestCase):
    @unittest.skipIf(is_windows, "Posix specific behavior")
    def test_catch_permission_error(self):
        root_path = self.make_path("rootpath")
        self.create_dir(root_path)
        dir1_path = self.os.path.join(root_path, "dir1")
        dir2_path = self.os.path.join(root_path, "dir2")
        self.create_dir(dir1_path)
        self.os.chmod(dir1_path, 0o555)  # remove write permissions
        self.create_dir(dir2_path)
        old_file_path = self.os.path.join(dir2_path, "f1.txt")
        new_file_path = self.os.path.join(dir1_path, "f1.txt")
        self.create_file(old_file_path)

        with self.assertRaises(PermissionError):
            shutil.move(old_file_path, new_file_path)

    def test_rmtree(self):
        directory = self.make_path("xyzzy")
        dir_path = os.path.join(directory, "subdir")
        self.create_dir(dir_path)
        file_path = os.path.join(directory, "subfile")
        self.create_file(file_path)
        self.assertTrue(os.path.exists(directory))
        shutil.rmtree(directory)
        self.assertFalse(os.path.exists(directory))
        self.assertFalse(os.path.exists(dir_path))
        self.assertFalse(os.path.exists(file_path))

    def test_rmtree_with_trailing_slash(self):
        directory = self.make_path("xyzzy")
        dir_path = os.path.join(directory, "subdir")
        self.create_dir(dir_path)
        file_path = os.path.join(directory, "subfile")
        self.create_file(file_path)
        shutil.rmtree(directory + "/")
        self.assertFalse(os.path.exists(directory))
        self.assertFalse(os.path.exists(dir_path))
        self.assertFalse(os.path.exists(file_path))

    @unittest.skipIf(not is_windows, "Windows specific behavior")
    def test_rmtree_without_permission_for_a_file_in_windows(self):
        self.check_windows_only()
        dir_path = self.make_path("foo")
        self.create_file(os.path.join(dir_path, "bar"))
        file_path = os.path.join(dir_path, "baz")
        self.create_file(file_path)
        self.os.chmod(file_path, 0o444)
        with self.assertRaises(OSError):
            shutil.rmtree(dir_path)
        self.assertTrue(os.path.exists(file_path))
        self.os.chmod(file_path, 0o666)

    @unittest.skipIf(is_windows, "Posix specific behavior")
    def test_rmtree_without_permission_for_a_dir_in_posix(self):
        self.check_posix_only()
        dir_path = self.make_path("foo")
        self.create_file(os.path.join(dir_path, "bar"))
        file_path = os.path.join(dir_path, "baz")
        self.create_file(file_path)
        self.os.chmod(dir_path, 0o555)
        if not is_root():
            with self.assertRaises(OSError):
                shutil.rmtree(dir_path)
            self.assertTrue(os.path.exists(file_path))
            self.os.chmod(dir_path, 0o777)
        else:
            shutil.rmtree(dir_path)
            self.assertFalse(os.path.exists(file_path))

    @unittest.skipIf(is_windows, "Posix specific behavior")
    def test_rmtree_with_open_file_posix(self):
        self.check_posix_only()
        dir_path = self.make_path("foo")
        self.create_file(os.path.join(dir_path, "bar"))
        file_path = os.path.join(dir_path, "baz")
        self.create_file(file_path)
        with open(file_path, encoding="utf8"):
            shutil.rmtree(dir_path)
        self.assertFalse(os.path.exists(file_path))

    @unittest.skipIf(not is_windows, "Windows specific behavior")
    def test_rmtree_with_open_file_fails_under_windows(self):
        self.check_windows_only()
        dir_path = self.make_path("foo")
        self.create_file(os.path.join(dir_path, "bar"))
        file_path = os.path.join(dir_path, "baz")
        self.create_file(file_path)
        with open(file_path, encoding="utf8"):
            with self.assertRaises(OSError):
                shutil.rmtree(dir_path)
        self.assertTrue(os.path.exists(dir_path))

    def test_rmtree_non_existing_dir(self):
        directory = "nonexisting"
        with self.assertRaises(OSError):
            shutil.rmtree(directory)
        try:
            shutil.rmtree(directory, ignore_errors=True)
        except OSError:
            self.fail("rmtree raised despite ignore_errors True")

    def test_rmtree_non_existing_dir_with_handler(self):
        class NonLocal:
            pass

        def error_handler(_, path, _error_info):
            NonLocal.errorHandled = True
            NonLocal.errorPath = path

        directory = self.make_path("nonexisting")
        NonLocal.errorHandled = False
        NonLocal.errorPath = ""
        try:
            shutil.rmtree(directory, onerror=error_handler)
        except OSError:
            self.fail("rmtree raised exception despite onerror defined")
        self.assertTrue(NonLocal.errorHandled)
        self.assertEqual(NonLocal.errorPath, directory)

        NonLocal.errorHandled = False
        NonLocal.errorPath = ""
        try:
            shutil.rmtree(directory, ignore_errors=True, onerror=error_handler)
        except OSError:
            self.fail("rmtree raised exception despite ignore_errors True")
        # ignore_errors is True, so the onerror() error handler was
        # not executed
        self.assertFalse(NonLocal.errorHandled)
        self.assertEqual(NonLocal.errorPath, "")

    def test_rmtree_in_windows(self):
        # regression test for #979
        self.check_windows_only()
        base_path = self.make_path("foo", "bar")
        self.os.makedirs(self.os.path.join(base_path, "res"))
        self.assertTrue(self.os.path.exists(base_path))
        shutil.rmtree(base_path)
        self.assertFalse(self.os.path.exists(base_path))

    def test_copy(self):
        src_file = self.make_path("xyzzy")
        dst_file = self.make_path("xyzzy_copy")
        self.create_file(src_file)
        os.chmod(src_file, 0o750)
        self.assertTrue(os.path.exists(src_file))
        self.assertFalse(os.path.exists(dst_file))
        shutil.copy(src_file, dst_file)
        self.assertTrue(os.path.exists(dst_file))
        self.assertEqual(os.stat(src_file).st_mode, os.stat(dst_file).st_mode)

    def test_copy_directory(self):
        src_file = self.make_path("xyzzy")
        parent_directory = self.make_path("parent")
        dst_file = os.path.join(parent_directory, "xyzzy")
        self.create_file(src_file)
        self.create_dir(parent_directory)
        os.chmod(src_file, 0o750)
        self.assertTrue(os.path.exists(src_file))
        self.assertTrue(os.path.exists(parent_directory))
        self.assertFalse(os.path.exists(dst_file))
        shutil.copy(src_file, parent_directory)
        self.assertTrue(os.path.exists(dst_file))
        self.assertEqual(os.stat(src_file).st_mode, os.stat(dst_file).st_mode)

    def test_copystat(self):
        src_file = self.make_path("xyzzy")
        self.create_file(src_file)
        os.chmod(src_file, 0o750)
        dst_file = self.make_path("xyzzy_copy")
        self.create_file(dst_file)
        self.assertTrue(os.path.exists(src_file))
        self.assertTrue(os.path.exists(dst_file))
        shutil.copystat(src_file, dst_file)
        src_stat = os.stat(src_file)
        dst_stat = os.stat(dst_file)
        self.assertEqual(src_stat.st_mode, dst_stat.st_mode)
        self.assertAlmostEqual(src_stat.st_atime, dst_stat.st_atime, places=0)
        self.assertAlmostEqual(src_stat.st_mtime, dst_stat.st_mtime, places=2)

    @unittest.skipIf(IS_PYPY, "Functionality not supported in PyPy")
    def test_copystat_symlinks(self):
        """Regression test for #799"""
        self.skip_if_symlink_not_supported()
        f = self.make_path("xyzzy")
        self.create_file(f)
        sym1 = self.make_path("sym1")
        sym2 = self.make_path("sym2")
        self.create_symlink(sym1, f)
        self.create_symlink(sym2, f)
        shutil.copystat(sym1, sym2, follow_symlinks=False)

    def test_copy2(self):
        src_file = self.make_path("xyzzy")
        self.create_file(src_file)
        os.chmod(src_file, 0o750)
        dst_file = self.make_path("xyzzy_copy")
        self.assertTrue(os.path.exists(src_file))
        self.assertFalse(os.path.exists(dst_file))
        shutil.copy2(src_file, dst_file)
        self.assertTrue(os.path.exists(dst_file))
        src_stat = os.stat(src_file)
        dst_stat = os.stat(dst_file)
        self.assertEqual(src_stat.st_mode, dst_stat.st_mode)
        self.assertAlmostEqual(src_stat.st_atime, dst_stat.st_atime, places=0)
        self.assertAlmostEqual(src_stat.st_mtime, dst_stat.st_mtime, places=2)

    def test_copy2_directory(self):
        src_file = self.make_path("xyzzy")
        parent_directory = self.make_path("parent")
        dst_file = os.path.join(parent_directory, "xyzzy")
        self.create_file(src_file)
        self.create_dir(parent_directory)
        os.chmod(src_file, 0o750)
        self.assertTrue(os.path.exists(src_file))
        self.assertTrue(os.path.exists(parent_directory))
        self.assertFalse(os.path.exists(dst_file))
        shutil.copy2(src_file, parent_directory)
        self.assertTrue(os.path.exists(dst_file))
        src_stat = os.stat(src_file)
        dst_stat = os.stat(dst_file)
        self.assertEqual(src_stat.st_mode, dst_stat.st_mode)
        self.assertAlmostEqual(src_stat.st_atime, dst_stat.st_atime, places=0)
        self.assertAlmostEqual(src_stat.st_mtime, dst_stat.st_mtime, places=2)

    def test_copytree(self):
        src_directory = self.make_path("xyzzy")
        dst_directory = self.make_path("xyzzy_copy")
        self.create_dir(src_directory)
        self.create_dir("%s/subdir" % src_directory)
        self.create_file(os.path.join(src_directory, "subfile"))
        self.assertTrue(os.path.exists(src_directory))
        self.assertFalse(os.path.exists(dst_directory))
        shutil.copytree(src_directory, dst_directory)
        self.assertTrue(os.path.exists(dst_directory))
        self.assertTrue(os.path.exists(os.path.join(dst_directory, "subdir")))
        self.assertTrue(os.path.exists(os.path.join(dst_directory, "subfile")))

    def test_copytree_src_is_file(self):
        src_file = self.make_path("xyzzy")
        dst_directory = self.make_path("xyzzy_copy")
        self.create_file(src_file)
        self.assertTrue(os.path.exists(src_file))
        self.assertFalse(os.path.exists(dst_directory))
        with self.assertRaises(OSError):
            shutil.copytree(src_file, dst_directory)

    def test_move_file_in_same_filesystem(self):
        self.skip_real_fs()
        src_file = "/original_xyzzy"
        dst_file = "/moved_xyzzy"
        src_object = self.fs.create_file(src_file)
        src_ino = src_object.st_ino
        src_dev = src_object.st_dev

        self.assertTrue(os.path.exists(src_file))
        self.assertFalse(os.path.exists(dst_file))
        shutil.move(src_file, dst_file)
        self.assertTrue(os.path.exists(dst_file))
        self.assertFalse(os.path.exists(src_file))

        dst_object = self.fs.get_object(dst_file)
        self.assertEqual(src_ino, dst_object.st_ino)
        self.assertEqual(src_dev, dst_object.st_dev)

    def test_move_file_into_other_filesystem(self):
        self.skip_real_fs()
        mount_point = self.create_mount_point()

        src_file = self.make_path("original_xyzzy")
        dst_file = self.os.path.join(mount_point, "moved_xyzzy")
        src_object = self.fs.create_file(src_file)
        src_ino = src_object.st_ino
        src_dev = src_object.st_dev

        shutil.move(src_file, dst_file)
        self.assertTrue(os.path.exists(dst_file))
        self.assertFalse(os.path.exists(src_file))

        dst_object = self.fs.get_object(dst_file)
        self.assertNotEqual(src_ino, dst_object.st_ino)
        self.assertNotEqual(src_dev, dst_object.st_dev)

    def test_move_file_into_directory(self):
        src_file = self.make_path("xyzzy")
        dst_directory = self.make_path("directory")
        dst_file = os.path.join(dst_directory, "xyzzy")
        self.create_file(src_file)
        self.create_dir(dst_directory)
        self.assertTrue(os.path.exists(src_file))
        self.assertFalse(os.path.exists(dst_file))
        shutil.move(src_file, dst_directory)
        self.assertTrue(os.path.exists(dst_file))
        self.assertFalse(os.path.exists(src_file))

    def test_move_directory(self):
        src_directory = self.make_path("original_xyzzy")
        dst_directory = self.make_path("moved_xyzzy")
        self.create_dir(src_directory)
        self.create_file(os.path.join(src_directory, "subfile"))
        self.create_dir(os.path.join(src_directory, "subdir"))
        self.assertTrue(os.path.exists(src_directory))
        self.assertFalse(os.path.exists(dst_directory))
        shutil.move(src_directory, dst_directory)
        self.assertTrue(os.path.exists(dst_directory))
        self.assertTrue(os.path.exists(os.path.join(dst_directory, "subfile")))
        self.assertTrue(os.path.exists(os.path.join(dst_directory, "subdir")))
        self.assertFalse(os.path.exists(src_directory))

    def test_disk_usage(self):
        self.skip_real_fs()
        file_path = self.make_path("foo", "bar")
        self.fs.create_file(file_path, st_size=400)
        disk_usage = shutil.disk_usage(file_path)
        self.assertEqual(1000, disk_usage.total)
        self.assertEqual(400, disk_usage.used)
        self.assertEqual(600, disk_usage.free)
        self.assertEqual((1000, 400, 600), disk_usage)

        mount_point = self.create_mount_point()
        dir_path = self.os.path.join(mount_point, "foo")
        file_path = self.os.path.join(dir_path, "bar")
        self.fs.create_file(file_path, st_size=400)
        disk_usage = shutil.disk_usage(dir_path)
        self.assertEqual((500, 400, 100), disk_usage)

    def test_disk_usage_with_path(self):
        self.skip_real_fs()
        file_path = self.make_path("foo", "bar")
        self.fs.create_file(file_path, st_size=400)
        path = Path(file_path)
        disk_usage = shutil.disk_usage(path)
        self.assertEqual(1000, disk_usage.total)
        self.assertEqual(400, disk_usage.used)
        self.assertEqual(600, disk_usage.free)
        self.assertEqual((1000, 400, 600), disk_usage)

    def create_mount_point(self):
        mount_point = "M:" if self.is_windows_fs else "/mount"
        self.fs.add_mount_point(mount_point, total_size=500)
        return mount_point


class RealShutilModuleTest(FakeShutilModuleTest):
    def use_real_fs(self):
        return True


class FakeCopyFileTest(RealFsTestCase):
    def tearDown(self):
        super().tearDown()

    def test_common_case(self):
        src_file = self.make_path("xyzzy")
        dst_file = self.make_path("xyzzy_copy")
        contents = "contents of file"
        self.create_file(src_file, contents=contents)
        self.assertTrue(os.path.exists(src_file))
        self.assertFalse(os.path.exists(dst_file))
        shutil.copyfile(src_file, dst_file)
        self.assertTrue(os.path.exists(dst_file))
        self.check_contents(dst_file, contents)

    def test_raises_if_source_and_dest_are_the_same_file(self):
        src_file = self.make_path("xyzzy")
        dst_file = src_file
        contents = "contents of file"
        self.create_file(src_file, contents=contents)
        self.assertTrue(os.path.exists(src_file))
        with self.assertRaises(shutil.Error):
            shutil.copyfile(src_file, dst_file)

    def test_raises_if_dest_is_a_symlink_to_src(self):
        self.skip_if_symlink_not_supported()
        src_file = self.make_path("foo")
        dst_file = self.make_path("bar")
        contents = "contents of file"
        self.create_file(src_file, contents=contents)
        self.create_symlink(dst_file, src_file)
        self.assertTrue(os.path.exists(src_file))
        with self.assertRaises(shutil.Error):
            shutil.copyfile(src_file, dst_file)

    def test_succeeds_if_dest_exists_and_is_writable(self):
        src_file = self.make_path("xyzzy")
        dst_file = self.make_path("xyzzy_copy")
        src_contents = "contents of source file"
        dst_contents = "contents of dest file"
        self.create_file(src_file, contents=src_contents)
        self.create_file(dst_file, contents=dst_contents)
        self.assertTrue(os.path.exists(src_file))
        self.assertTrue(os.path.exists(dst_file))
        shutil.copyfile(src_file, dst_file)
        self.assertTrue(os.path.exists(dst_file))
        self.check_contents(dst_file, src_contents)

    def test_raises_if_dest_exists_and_is_not_writable(self):
        src_file = self.make_path("xyzzy")
        dst_file = self.make_path("xyzzy_copy")
        src_contents = "contents of source file"
        dst_contents = "contents of dest file"
        self.create_file(src_file, contents=src_contents)
        self.create_file(dst_file, contents=dst_contents)
        os.chmod(dst_file, 0o400)
        self.assertTrue(os.path.exists(src_file))
        self.assertTrue(os.path.exists(dst_file))

        if is_root():
            shutil.copyfile(src_file, dst_file)
            self.assertTrue(self.os.path.exists(dst_file))
            with self.open(dst_file) as f:
                self.assertEqual("contents of source file", f.read())
        else:
            with self.assertRaises(OSError):
                shutil.copyfile(src_file, dst_file)

        os.chmod(dst_file, 0o666)

    @unittest.skipIf(is_windows, "Posix specific behavior")
    def test_raises_if_dest_dir_is_not_writable_under_posix(self):
        self.check_posix_only()
        src_file = self.make_path("xyzzy")
        dst_dir = self.make_path("tmp", "foo")
        dst_file = os.path.join(dst_dir, "xyzzy")
        src_contents = "contents of source file"
        self.create_file(src_file, contents=src_contents)
        self.create_dir(dst_dir)
        os.chmod(dst_dir, 0o555)
        self.assertTrue(os.path.exists(src_file))
        self.assertTrue(os.path.exists(dst_dir))
        if not is_root():
            with self.assertRaises(OSError):
                shutil.copyfile(src_file, dst_file)
        else:
            shutil.copyfile(src_file, dst_file)
            self.assertTrue(os.path.exists(dst_file))
            self.check_contents(dst_file, src_contents)

    def test_raises_if_src_doesnt_exist(self):
        src_file = self.make_path("xyzzy")
        dst_file = self.make_path("xyzzy_copy")
        self.assertFalse(os.path.exists(src_file))
        with self.assertRaises(OSError):
            shutil.copyfile(src_file, dst_file)

    @unittest.skipIf(is_windows, "Posix specific behavior")
    def test_raises_if_src_not_readable(self):
        self.check_posix_only()
        src_file = self.make_path("xyzzy")
        dst_file = self.make_path("xyzzy_copy")
        src_contents = "contents of source file"
        self.create_file(src_file, contents=src_contents)
        os.chmod(src_file, 0o000)
        self.assertTrue(os.path.exists(src_file))
        if not is_root():
            with self.assertRaises(OSError):
                shutil.copyfile(src_file, dst_file)
        else:
            shutil.copyfile(src_file, dst_file)
            self.assertTrue(os.path.exists(dst_file))
            self.check_contents(dst_file, src_contents)

    def test_raises_if_src_is_a_directory(self):
        src_file = self.make_path("xyzzy")
        dst_file = self.make_path("xyzzy_copy")
        self.create_dir(src_file)
        self.assertTrue(os.path.exists(src_file))
        with self.assertRaises(OSError):
            shutil.copyfile(src_file, dst_file)

    def test_raises_if_dest_is_a_directory(self):
        src_file = self.make_path("xyzzy")
        dst_dir = self.make_path("tmp", "foo")
        src_contents = "contents of source file"
        self.create_file(src_file, contents=src_contents)
        self.create_dir(dst_dir)
        self.assertTrue(os.path.exists(src_file))
        self.assertTrue(os.path.exists(dst_dir))
        with self.assertRaises(OSError):
            shutil.copyfile(src_file, dst_dir)

    def test_moving_dir_into_dir(self):
        # regression test for #515
        source_dir = tempfile.mkdtemp()
        target_dir = tempfile.mkdtemp()
        filename = "foo.pdf"
        with open(os.path.join(source_dir, filename), "wb") as fp:
            fp.write(b"stub")

        shutil.move(source_dir, target_dir)
        shutil.rmtree(target_dir)


class RealCopyFileTest(FakeCopyFileTest):
    def use_real_fs(self):
        return True


if __name__ == "__main__":
    unittest.main()



# Copyright 2009 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A fake shutil module implementation that uses fake_filesystem for
unit tests.
Note that only `shutildisk_usage()` is faked, the rest of the functions shall
work fine with the fake file system if `os`/`os.path` are patched.

:Includes:
  FakeShutil: Uses a FakeFilesystem to provide a fake replacement for the
    shutil module.

:Usage:
  The fake implementation is automatically involved if using
  `fake_filesystem_unittest.TestCase`, pytest fs fixture,
  or directly `Patcher`.
"""

import os
import shutil
import sys


class FakeShutilModule:
    """Uses a FakeFilesystem to provide a fake replacement
    for shutil module.
    """

    @staticmethod
    def dir():
        """Return the list of patched function names. Used for patching
        functions imported from the module.
        """
        return ("disk_usage",)

    def __init__(self, filesystem):
        """Construct fake shutil module using the fake filesystem.

        Args:
          filesystem:  FakeFilesystem used to provide file system information
        """
        self.filesystem = filesystem
        self._shutil_module = shutil

    def disk_usage(self, path):
        """Return the total, used and free disk space in bytes as named tuple
        or placeholder holder values simulating unlimited space if not set.

        Args:
          path: defines the filesystem device which is queried
        """
        return self.filesystem.get_disk_usage(path)

    if sys.version_info >= (3, 12) and sys.platform == "win32":

        def copy2(self, src, dst, *, follow_symlinks=True):
            """Since Python 3.12, there is an optimization fow Windows,
            using the Windows API. We just remove this and fall back to the previous
            implementation.
            """
            if self.filesystem.isdir(dst):
                dst = self.filesystem.joinpaths(dst, os.path.basename(src))

            self.copyfile(src, dst, follow_symlinks=follow_symlinks)
            self.copystat(src, dst, follow_symlinks=follow_symlinks)
            return dst

        def copytree(
            self,
            src,
            dst,
            symlinks=False,
            ignore=None,
            copy_function=shutil.copy2,
            ignore_dangling_symlinks=False,
            dirs_exist_ok=False,
        ):
            """Make sure the default argument is patched."""
            if copy_function == shutil.copy2:
                copy_function = self.copy2
            return self._shutil_module.copytree(
                src,
                dst,
                symlinks,
                ignore,
                copy_function,
                ignore_dangling_symlinks,
                dirs_exist_ok,
            )

        def move(self, src, dst, copy_function=shutil.copy2):
            """Make sure the default argument is patched."""
            if copy_function == shutil.copy2:
                copy_function = self.copy2
            return self._shutil_module.move(src, dst, copy_function)

    def __getattr__(self, name):
        """Forwards any non-faked calls to the standard shutil module."""
        return getattr(self._shutil_module, name)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides functions for testing additional_skip_names functionality.
"""

import os
from pathlib import Path


def read_pathlib(file_name):
    return (Path(__file__).parent / file_name).open("r").read()


def read_text_pathlib(file_name):
    return (Path(__file__).parent / file_name).read_text()


def read_bytes_pathlib(file_name):
    return (Path(__file__).parent / file_name).read_bytes()


def check_exists_pathlib():
    return os.path.exists(__file__) and Path(__file__).exists()


def read_open(file_name):
    with open(os.path.join(os.path.dirname(__file__), file_name)) as f:
        return f.read()



# Copyright 2009 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test that FakeFilesystem calls work identically to a real filesystem."""
# pylint: disable-all

import os
import shutil
import sys
import tempfile
import time
import unittest

from pyfakefs import fake_filesystem, fake_os, fake_open
from pyfakefs.helpers import IS_PYPY


def sep(path):
    """Converts slashes in the path to the architecture's path separator."""
    if isinstance(path, str):
        return path.replace("/", os.sep)
    return path


def _get_errno(raised_error):
    if raised_error is not None:
        try:
            return raised_error.errno
        except AttributeError:
            pass


class TestCase(unittest.TestCase):
    is_windows = sys.platform.startswith("win")
    _FAKE_FS_BASE = "C:\\fakefs" if is_windows else "/fakefs"


class FakeFilesystemVsRealTest(TestCase):
    def _paths(self, path):
        """For a given path, return paths in the real and fake filesystems."""
        if not path:
            return None, None
        return (
            os.path.join(self.real_base, path),
            os.path.join(self.fake_base, path),
        )

    def _create_test_file(self, file_type, path, contents=None):
        """Create a dir, file, or link in both the real fs and the fake."""
        path = sep(path)
        self._created_files.append([file_type, path, contents])
        real_path, fake_path = self._paths(path)
        if file_type == "d":
            os.mkdir(real_path)
            self.fake_os.mkdir(fake_path)
        if file_type == "f":
            fh = open(real_path, "w", encoding="utf8")
            fh.write(contents or "")
            fh.close()
            fh = self.fake_open(fake_path, "w", encoding="utf8")
            fh.write(contents or "")
            fh.close()
        # b for binary file
        if file_type == "b":
            fh = open(real_path, "wb")
            fh.write(contents or "")
            fh.close()
            fh = self.fake_open(fake_path, "wb")
            fh.write(contents or "")
            fh.close()
        # l for symlink, h for hard link
        if file_type in ("l", "h"):
            real_target, fake_target = (contents, contents)
            # If it begins with '/', make it relative to the base. You can't go
            # creating files in / for the real file system.
            if contents.startswith(os.sep):
                real_target, fake_target = self._paths(contents[1:])
            if file_type == "l":
                os.symlink(real_target, real_path)
                self.fake_os.symlink(fake_target, fake_path)
            elif file_type == "h":
                os.link(real_target, real_path)
                self.fake_os.link(fake_target, fake_path)

    def setUp(self):
        # Base paths in the real and test file systems. We keep them different
        # so that missing features in the fake don't fall through to the base
        # operations and magically succeed.
        tsname = "fakefs.%s" % time.time()
        self.cwd = os.getcwd()
        # Fully expand the base_path - required on OS X.
        self.real_base = os.path.realpath(os.path.join(tempfile.gettempdir(), tsname))
        os.chdir(tempfile.gettempdir())
        if os.path.isdir(self.real_base):
            shutil.rmtree(self.real_base)
        os.mkdir(self.real_base)
        self.fake_base = self._FAKE_FS_BASE

        # Make sure we can write to the physical testing temp directory.
        self.assertTrue(os.access(self.real_base, os.W_OK))

        self.fake_filesystem = fake_filesystem.FakeFilesystem()
        self.fake_filesystem.create_dir(self.fake_base)
        self.fake_os = fake_os.FakeOsModule(self.fake_filesystem)
        self.fake_open = fake_open.FakeFileOpen(self.fake_filesystem)
        self._created_files = []

        os.chdir(self.real_base)
        self.fake_os.chdir(self.fake_base)

    def tearDown(self):
        # We have to remove all the files from the real FS. Doing the same for
        # the fake FS is optional, but doing it is an extra sanity check.
        os.chdir(tempfile.gettempdir())
        try:
            rev_files = self._created_files[:]
            rev_files.reverse()
            for info in rev_files:
                real_path, fake_path = self._paths(info[1])
                if info[0] == "d":
                    try:
                        os.rmdir(real_path)
                    except OSError as e:
                        if "Directory not empty" in e:
                            self.fail(
                                "Real path %s not empty: %s : %s"
                                % (real_path, e, os.listdir(real_path))
                            )
                        else:
                            raise
                    self.fake_os.rmdir(fake_path)
                if info[0] == "f" or info[0] == "l":
                    os.remove(real_path)
                    self.fake_os.remove(fake_path)
        finally:
            shutil.rmtree(self.real_base)
            os.chdir(self.cwd)

    def _compare_behaviors(
        self, method_name, path, real, fake, method_returns_path=False
    ):
        """Invoke an os method in both real and fake contexts and compare
        results.

        Invoke a real filesystem method with a path to a real file and invoke
        a fake filesystem method with a path to a fake file and compare the
        results. We expect some calls to throw Exceptions, so we catch those
        and compare them.

        Args:
            method_name: Name of method being tested, for use in
                error messages.
            path: potential path to a file in the real and fake file systems,
                passing an empty tuple indicates that no arguments to pass
                to method.
            real: built-in system library or method from the built-in system
                library which takes a path as an arg and returns some value.
            fake: fake_filsystem object or method from a fake_filesystem class
                which takes a path as an arg and returns some value.
            method_returns_path: True if the method returns a path, and thus we
                must compensate for expected difference between real and fake.

        Returns:
            A description of the difference in behavior, or None.
        """
        # pylint: disable=C6403

        def _error_class(exc):
            if exc:
                if hasattr(exc, "errno"):
                    return f"{exc.__class__.__name__}({exc.errno})"
                return exc.__class__.__name__
            return "None"

        real_err, real_value = self._get_real_value(method_name, path, real)
        fake_err, fake_value = self._get_fake_value(method_name, path, fake)

        method_call = f"{method_name}"
        method_call += "()" if path == () else "({path})"
        # We only compare on the error class because the actual error contents
        # is almost always different because of the file paths.
        if _error_class(real_err) != _error_class(fake_err):
            if real_err is None:
                return "{}: real version returned {}, fake raised {}".format(
                    method_call,
                    real_value,
                    _error_class(fake_err),
                )
            if fake_err is None:
                return "{}: real version raised {}, fake returned {}".format(
                    method_call,
                    _error_class(real_err),
                    fake_value,
                )
            return "{}: real version raised {}, fake raised {}".format(
                method_call,
                _error_class(real_err),
                _error_class(fake_err),
            )
        real_errno = _get_errno(real_err)
        fake_errno = _get_errno(fake_err)
        if real_errno != fake_errno:
            return "{}({}): both raised {}, real errno {}, fake errno {}".format(
                method_name,
                path,
                _error_class(real_err),
                real_errno,
                fake_errno,
            )
        # If the method is supposed to return a full path AND both values
        # begin with the expected full path, then trim it off.
        if method_returns_path:
            if (
                real_value
                and fake_value
                and real_value.startswith(self.real_base)
                and fake_value.startswith(self.fake_base)
            ):
                real_value = real_value[len(self.real_base) :]
                fake_value = fake_value[len(self.fake_base) :]
        if real_value != fake_value:
            return "{}: real return {}, fake returned {}".format(
                method_call,
                real_value,
                fake_value,
            )
        return None

    @staticmethod
    def _get_fake_value(method_name, path, fake):
        fake_value = None
        fake_err = None
        try:
            fake_method = fake
            if not callable(fake):
                fake_method = getattr(fake, method_name)
            args = [] if path == () else [path]
            result = fake_method(*args)
            if isinstance(result, bytes):
                fake_value = result.decode()
            else:
                fake_value = str(result)
        except Exception as e:  # pylint: disable-msg=W0703
            fake_err = e
        return fake_err, fake_value

    @staticmethod
    def _get_real_value(method_name, path, real):
        real_value = None
        real_err = None
        # Catching Exception below gives a lint warning, but it's what we need.
        try:
            args = [] if path == () else [path]
            real_method = real
            if not callable(real):
                real_method = getattr(real, method_name)
            result = real_method(*args)
            if isinstance(result, bytes):
                real_value = result.decode()
            else:
                real_value = str(result)
        except Exception as e:  # pylint: disable-msg=W0703
            real_err = e
        return real_err, real_value

    def assertOsMethodBehaviorMatches(
        self, method_name, path, method_returns_path=False
    ):
        """Invoke an os method in both real and fake contexts and compare.

        For a given method name (from the os module) and a path, compare the
        behavior of the system provided module against the fake_filesystem
        module.
        We expect results and/or Exceptions raised to be identical.

        Args:
            method_name: Name of method being tested.
            path: potential path to a file in the real and fake file systems.
            method_returns_path: True if the method returns a path, and thus we
                must compensate for expected difference between real and fake.

        Returns:
            A description of the difference in behavior, or None.
        """
        path = sep(path)
        return self._compare_behaviors(
            method_name, path, os, self.fake_os, method_returns_path
        )

    def diff_open_method_behavior(
        self, method_name, path, mode, data, method_returns_data=True
    ):
        """Invoke an open method in both real and fkae contexts and compare.

        Args:
            method_name: Name of method being tested.
            path: potential path to a file in the real and fake file systems.
            mode: how to open the file.
            data: any data to pass to the method.
            method_returns_data: True if a method returns some sort of data.

        For a given method name (from builtin open) and a path, compare the
        behavior of the system provided module against the fake_filesystem
        module.
        We expect results and/or Exceptions raised to be identical.

        Returns:
            A description of the difference in behavior, or None.
        """
        kwargs = {}
        if "b" not in mode:
            kwargs["encoding"] = "utf8"
        with open(path, mode, **kwargs) as real_fh:
            with self.fake_open(path, mode, **kwargs) as fake_fh:
                return self._compare_behaviors(
                    method_name, data, real_fh, fake_fh, method_returns_data
                )

    def diff_os_path_method_behavior(
        self, method_name, path, method_returns_path=False
    ):
        """Invoke an os.path method in both real and fake contexts and compare.

        For a given method name (from the os.path module) and a path, compare
        the behavior of the system provided module against the
        fake_filesytem module.
        We expect results and/or Exceptions raised to be identical.

        Args:
            method_name: Name of method being tested.
            path: potential path to a file in the real and fake file systems.
            method_returns_path: True if the method returns a path, and thus we
                must compensate for expected difference between real and fake.

        Returns:
            A description of the difference in behavior, or None.
        """
        return self._compare_behaviors(
            method_name, path, os.path, self.fake_os.path, method_returns_path
        )

    def assertOsPathMethodBehaviorMatches(
        self, method_name, path, method_returns_path=False
    ):
        """Assert that an os.path behaves the same in both real and
        fake contexts.

        Wraps DiffOsPathMethodBehavior, raising AssertionError if any
        differences are reported.

        Args:
            method_name: Name of method being tested.
            path: potential path to a file in the real and fake file systems.
            method_returns_path: True if the method returns a path, and thus we
                must compensate for expected difference between real and fake.

        Raises:
            AssertionError if there is any difference in behavior.
        """
        path = sep(path)
        diff = self.diff_os_path_method_behavior(method_name, path, method_returns_path)
        if diff:
            self.fail(diff)

    def assertAllOsBehaviorsMatch(self, path):
        path = sep(path)
        os_method_names = [] if self.is_windows else ["readlink"]
        os_method_names_no_args = ["getcwd"]
        os_path_method_names = ["isabs", "isdir"]
        if not self.is_windows:
            os_path_method_names += ["islink", "lexists"]
        if not self.is_windows or not IS_PYPY:
            os_path_method_names += ["isfile", "exists"]

        wrapped_methods = [
            ["access", self._access_real, self._access_fake],
            ["stat.size", self._stat_size_real, self._stat_size_fake],
            ["lstat.size", self._lstat_size_real, self._lstat_size_fake],
        ]

        differences = []
        for method_name in os_method_names:
            diff = self.assertOsMethodBehaviorMatches(method_name, path)
            if diff:
                differences.append(diff)
        for method_name in os_method_names_no_args:
            diff = self.assertOsMethodBehaviorMatches(
                method_name, (), method_returns_path=True
            )
            if diff:
                differences.append(diff)
        for method_name in os_path_method_names:
            diff = self.diff_os_path_method_behavior(method_name, path)
            if diff:
                differences.append(diff)
        for m in wrapped_methods:
            diff = self._compare_behaviors(m[0], path, m[1], m[2])
            if diff:
                differences.append(diff)
        if differences:
            self.fail(
                "Behaviors do not match for %s:\n    %s"
                % (path, "\n    ".join(differences))
            )

    def assertFileHandleBehaviorsMatch(self, path, mode, data):
        path = sep(path)
        write_method_names = ["write", "writelines"]
        read_method_names = ["read", "readlines"]
        other_method_names = ["truncate", "flush", "close"]
        differences = []
        for method_name in write_method_names:
            diff = self.diff_open_method_behavior(method_name, path, mode, data)
            if diff:
                differences.append(diff)
        for method_name in read_method_names + other_method_names:
            diff = self.diff_open_method_behavior(method_name, path, mode, ())
            if diff:
                differences.append(diff)
        if differences:
            self.fail(
                "Behaviors do not match for %s:\n    %s"
                % (path, "\n    ".join(differences))
            )

    def assertFileHandleOpenBehaviorsMatch(self, *args, **kwargs):
        """Compare open() function invocation between real and fake.

        Runs open(*args, **kwargs) on both real and fake.

        Args:
            *args: args to pass through to open()
            **kwargs: kwargs to pass through to open().

        Returns:
            None.

        Raises:
            AssertionError if underlying open() behavior differs from fake.
        """
        real_err = None
        fake_err = None
        try:
            with open(*args, **kwargs):
                pass
        except Exception as e:  # pylint: disable-msg=W0703
            real_err = e

        try:
            with self.fake_open(*args, **kwargs):
                pass
        except Exception as e:  # pylint: disable-msg=W0703
            fake_err = e

        # default equal in case one is None and other is not.
        is_exception_equal = real_err == fake_err
        if real_err and fake_err:
            # exception __eq__ doesn't evaluate equal ever, thus manual check.
            is_exception_equal = (
                type(real_err) is type(fake_err) and real_err.args == fake_err.args
            )

        if not is_exception_equal:
            msg = "Behaviors don't match on open with args {} & kwargs {}.\n".format(
                args,
                kwargs,
            )
            real_err_msg = "Real open results in: %s\n" % repr(real_err)
            fake_err_msg = "Fake open results in: %s\n" % repr(fake_err)
            self.fail(msg + real_err_msg + fake_err_msg)

    # Helpers for checks which are not straight method calls.
    @staticmethod
    def _access_real(path):
        return os.access(path, os.R_OK)

    def _access_fake(self, path):
        return self.fake_os.access(path, os.R_OK)

    def _stat_size_real(self, path):
        real_path, unused_fake_path = self._paths(path)
        # fake_filesystem.py does not implement stat().st_size for directories
        if os.path.isdir(real_path):
            return None
        return os.stat(real_path).st_size

    def _stat_size_fake(self, path):
        unused_real_path, fake_path = self._paths(path)
        # fake_filesystem.py does not implement stat().st_size for directories
        if self.fake_os.path.isdir(fake_path):
            return None
        return self.fake_os.stat(fake_path).st_size

    def _lstat_size_real(self, path):
        real_path, unused_fake_path = self._paths(path)
        if os.path.isdir(real_path):
            return None
        size = os.lstat(real_path).st_size
        # Account for the difference in the lengths of the absolute paths.
        if os.path.islink(real_path):
            if os.readlink(real_path).startswith(os.sep):
                size -= len(self.real_base)
        return size

    def _lstat_size_fake(self, path):
        unused_real_path, fake_path = self._paths(path)
        # size = 0
        if self.fake_os.path.isdir(fake_path):
            return None
        size = self.fake_os.lstat(fake_path).st_size
        # Account for the difference in the lengths of the absolute paths.
        if self.fake_os.path.islink(fake_path):
            if self.fake_os.readlink(fake_path).startswith(os.sep):
                size -= len(self.fake_base)
        return size

    def test_isabs(self):
        # We do not have to create any files for isabs.
        self.assertOsPathMethodBehaviorMatches("isabs", None)
        self.assertOsPathMethodBehaviorMatches("isabs", "")
        self.assertOsPathMethodBehaviorMatches("isabs", "/")
        self.assertOsPathMethodBehaviorMatches("isabs", "/a")
        self.assertOsPathMethodBehaviorMatches("isabs", "a")

    def test_none_path(self):
        self.assertAllOsBehaviorsMatch(None)

    def test_empty_path(self):
        self.assertAllOsBehaviorsMatch("")

    def test_root_path(self):
        self.assertAllOsBehaviorsMatch("/")

    def test_non_existant_file(self):
        self.assertAllOsBehaviorsMatch("foo")

    def test_empty_file(self):
        self._create_test_file("f", "aFile")
        self.assertAllOsBehaviorsMatch("aFile")

    def test_file_with_contents(self):
        self._create_test_file("f", "aFile", "some contents")
        self.assertAllOsBehaviorsMatch("aFile")

    def test_file_with_binary_contents(self):
        self._create_test_file("b", "aFile", b"some contents")
        self.assertAllOsBehaviorsMatch("aFile")

    @unittest.skipIf(TestCase.is_windows, "no symlink in Windows")
    def test_sym_link_to_empty_file(self):
        self._create_test_file("f", "aFile")
        self._create_test_file("l", "link_to_empty", "aFile")
        self.assertAllOsBehaviorsMatch("link_to_empty")

    @unittest.skipIf(TestCase.is_windows, "no symlink in Windows")
    def test_hard_link_to_empty_file(self):
        self._create_test_file("f", "aFile")
        self._create_test_file("h", "link_to_empty", "aFile")
        self.assertAllOsBehaviorsMatch("link_to_empty")

    @unittest.skipIf(TestCase.is_windows, "no symlink in Windows")
    def test_sym_link_to_real_file(self):
        self._create_test_file("f", "aFile", "some contents")
        self._create_test_file("l", "link_to_file", "aFile")
        self.assertAllOsBehaviorsMatch("link_to_file")

    @unittest.skipIf(TestCase.is_windows, "no symlink in Windows")
    def test_hard_link_to_real_file(self):
        self._create_test_file("f", "aFile", "some contents")
        self._create_test_file("h", "link_to_file", "aFile")
        self.assertAllOsBehaviorsMatch("link_to_file")

    @unittest.skipIf(TestCase.is_windows, "no symlink in Windows")
    def test_broken_sym_link(self):
        self._create_test_file("l", "broken_link", "broken")
        self._create_test_file("l", "loop", "/a/loop")
        self.assertAllOsBehaviorsMatch("broken_link")

    def test_file_in_a_folder(self):
        self._create_test_file("d", "a")
        self._create_test_file("d", "a/b")
        self._create_test_file("f", "a/b/file", "contents")
        self.assertAllOsBehaviorsMatch("a/b/file")

    @unittest.skipIf(TestCase.is_windows, "no symlink in Windows")
    def test_absolute_sym_link_to_folder(self):
        self._create_test_file("d", "a")
        self._create_test_file("d", "a/b")
        self._create_test_file("f", "a/b/file", "contents")
        self._create_test_file("l", "a/link", "/a/b")
        self.assertAllOsBehaviorsMatch("a/link/file")

    @unittest.skipIf(TestCase.is_windows, "no symlink in Windows")
    def test_link_to_folder_after_chdir(self):
        self._create_test_file("d", "a")
        self._create_test_file("d", "a/b")
        self._create_test_file("f", "a/b/file", "contents")
        self._create_test_file("l", "a/link", "/a/b")

        real_dir, fake_dir = self._paths("a/b")
        os.chdir(real_dir)
        self.fake_os.chdir(fake_dir)
        self.assertAllOsBehaviorsMatch("file")

    @unittest.skipIf(TestCase.is_windows, "no symlink in Windows")
    def test_relative_sym_link_to_folder(self):
        self._create_test_file("d", "a")
        self._create_test_file("d", "a/b")
        self._create_test_file("f", "a/b/file", "contents")
        self._create_test_file("l", "a/link", "b")
        self.assertAllOsBehaviorsMatch("a/link/file")

    @unittest.skipIf(TestCase.is_windows, "no symlink in Windows")
    def test_sym_link_to_parent(self):
        # Soft links on HFS+ / OS X behave differently.
        if os.uname()[0] != "Darwin":
            self._create_test_file("d", "a")
            self._create_test_file("d", "a/b")
            self._create_test_file("l", "a/b/c", "..")
            self.assertAllOsBehaviorsMatch("a/b/c")

    @unittest.skipIf(TestCase.is_windows, "no symlink in Windows")
    def test_path_through_sym_link_to_parent(self):
        self._create_test_file("d", "a")
        self._create_test_file("f", "a/target", "contents")
        self._create_test_file("d", "a/b")
        self._create_test_file("l", "a/b/c", "..")
        self.assertAllOsBehaviorsMatch("a/b/c/target")

    @unittest.skipIf(TestCase.is_windows, "no symlink in Windows")
    def test_sym_link_to_sibling_directory(self):
        self._create_test_file("d", "a")
        self._create_test_file("d", "a/b")
        self._create_test_file("d", "a/sibling_of_b")
        self._create_test_file("f", "a/sibling_of_b/target", "contents")
        self._create_test_file("l", "a/b/c", "../sibling_of_b")
        self.assertAllOsBehaviorsMatch("a/b/c/target")

    @unittest.skipIf(TestCase.is_windows, "no symlink in Windows")
    def test_sym_link_to_sibling_directory_non_existant_file(self):
        self._create_test_file("d", "a")
        self._create_test_file("d", "a/b")
        self._create_test_file("d", "a/sibling_of_b")
        self._create_test_file("f", "a/sibling_of_b/target", "contents")
        self._create_test_file("l", "a/b/c", "../sibling_of_b")
        self.assertAllOsBehaviorsMatch("a/b/c/file_does_not_exist")

    @unittest.skipIf(TestCase.is_windows, "no symlink in Windows")
    def test_broken_sym_link_to_sibling_directory(self):
        self._create_test_file("d", "a")
        self._create_test_file("d", "a/b")
        self._create_test_file("d", "a/sibling_of_b")
        self._create_test_file("f", "a/sibling_of_b/target", "contents")
        self._create_test_file("l", "a/b/c", "../broken_sibling_of_b")
        self.assertAllOsBehaviorsMatch("a/b/c/target")

    def test_relative_path(self):
        self._create_test_file("d", "a")
        self._create_test_file("d", "a/b")
        self._create_test_file("d", "a/sibling_of_b")
        self._create_test_file("f", "a/sibling_of_b/target", "contents")
        self.assertAllOsBehaviorsMatch("a/b/../sibling_of_b/target")

    def test_broken_relative_path(self):
        self._create_test_file("d", "a")
        self._create_test_file("d", "a/b")
        self._create_test_file("d", "a/sibling_of_b")
        self._create_test_file("f", "a/sibling_of_b/target", "contents")
        self.assertAllOsBehaviorsMatch("a/b/../broken/target")

    def test_bad_relative_path(self):
        self._create_test_file("d", "a")
        self._create_test_file("f", "a/target", "contents")
        self._create_test_file("d", "a/b")
        self._create_test_file("d", "a/sibling_of_b")
        self._create_test_file("f", "a/sibling_of_b/target", "contents")
        self.assertAllOsBehaviorsMatch("a/b/../broken/../target")

    def test_getmtime_nonexistant_path(self):
        self.assertOsPathMethodBehaviorMatches("getmtime", "no/such/path")

    def test_builtin_open_modes(self):
        self._create_test_file("f", "read", "some contents")
        self._create_test_file("f", "write", "some contents")
        self._create_test_file("f", "append", "some contents")
        self.assertFileHandleBehaviorsMatch("read", "r", "other contents")
        self.assertFileHandleBehaviorsMatch("write", "w", "other contents")
        self.assertFileHandleBehaviorsMatch("append", "a", "other contents")
        self._create_test_file("f", "readplus", "some contents")
        self._create_test_file("f", "writeplus", "some contents")
        self.assertFileHandleBehaviorsMatch("readplus", "r+", "other contents")
        self.assertFileHandleBehaviorsMatch("writeplus", "w+", "other contents")
        self._create_test_file("b", "binaryread", b"some contents")
        self._create_test_file("b", "binarywrite", b"some contents")
        self._create_test_file("b", "binaryappend", b"some contents")
        self.assertFileHandleBehaviorsMatch("binaryread", "rb", b"other contents")
        self.assertFileHandleBehaviorsMatch("binarywrite", "wb", b"other contents")
        self.assertFileHandleBehaviorsMatch("binaryappend", "ab", b"other contents")
        self.assertFileHandleBehaviorsMatch("read", "rb", "other contents")
        self.assertFileHandleBehaviorsMatch("write", "wb", "other contents")
        self.assertFileHandleBehaviorsMatch("append", "ab", "other contents")

        # binary cannot have encoding
        self.assertFileHandleOpenBehaviorsMatch("read", "rb", encoding="enc")
        self.assertFileHandleOpenBehaviorsMatch("write", mode="wb", encoding="enc")
        self.assertFileHandleOpenBehaviorsMatch("append", "ab", encoding="enc")

        # text can have encoding
        self.assertFileHandleOpenBehaviorsMatch("read", "r", encoding="utf-8")
        self.assertFileHandleOpenBehaviorsMatch("write", "w", encoding="utf-8")
        self.assertFileHandleOpenBehaviorsMatch("append", "a", encoding="utf-8")


def main(_):
    unittest.main()


if __name__ == "__main__":
    unittest.main()



# Copyright 2009 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for glob using fake_filesystem."""

import contextlib
import glob
import os
import sys
import unittest

from pyfakefs import fake_filesystem_unittest


class FakeGlobUnitTest(fake_filesystem_unittest.TestCase):
    def setUp(self):
        self.setUpPyfakefs()
        directory = "./xyzzy"
        self.fs.create_dir(directory)
        self.fs.create_dir("%s/subdir" % directory)
        self.fs.create_dir("%s/subdir2" % directory)
        self.fs.create_file("%s/subfile" % directory)
        self.fs.create_file("[Temp]")

    def test_glob_empty(self):
        self.assertEqual(glob.glob(""), [])

    def test_glob_star(self):
        basedir = "/xyzzy"
        self.assertEqual(
            [
                os.path.join(basedir, "subdir"),
                os.path.join(basedir, "subdir2"),
                os.path.join(basedir, "subfile"),
            ],
            sorted(glob.glob("/xyzzy/*")),
        )

    def test_glob_exact(self):
        self.assertEqual(["/xyzzy"], glob.glob("/xyzzy"))
        self.assertEqual(["/xyzzy/subfile"], glob.glob("/xyzzy/subfile"))

    def test_glob_question(self):
        basedir = "/xyzzy"
        self.assertEqual(
            [
                os.path.join(basedir, "subdir"),
                os.path.join(basedir, "subdir2"),
                os.path.join(basedir, "subfile"),
            ],
            sorted(glob.glob("/x?zz?/*")),
        )

    def test_glob_no_magic(self):
        self.assertEqual(["/xyzzy"], glob.glob("/xyzzy"))
        self.assertEqual(["/xyzzy/subdir"], glob.glob("/xyzzy/subdir"))

    def test_non_existent_path(self):
        self.assertEqual([], glob.glob("nonexistent"))

    def test_magic_dir(self):
        self.assertEqual(["/[Temp]"], glob.glob("/*emp*"))

    def test_glob1(self):
        with (
            contextlib.nullcontext()
            if sys.version_info < (3, 13)
            else self.assertWarns(DeprecationWarning)
        ):
            self.assertEqual(["[Temp]"], glob.glob1("/", "*Tem*"))

    def test_has_magic(self):
        self.assertTrue(glob.has_magic("["))
        self.assertFalse(glob.has_magic("a"))


if __name__ == "__main__":
    unittest.main()
